# Providers information and getters/printers for inspection of the current logging setup
import logging


def all_loggers() -> list[logging.Logger]:
    loggers = [logging.getLogger()]  # get the root logger
    loggers = loggers + [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    return loggers


def print_all_loggers(print_fn: callable):
    loggers = all_loggers()
    printified_loggers_list = [_stringify_logger(logger) for logger in loggers]
    printified_loggers = _stringify_list(printified_loggers_list)
    print_fn(f"\n{printified_loggers}")


def _stringify_list(a: list, num_tabs: int=1) -> str:
    """ Concatanate a list of strings, placing each item on a newline. """
    if len(a) <= 0:
        return ''
    tabs = "\t" * num_tabs
    stringified = '\n'.join([f"{tabs}{list_item}" for list_item in a])
    return stringified

def _stringify_logger(logger: logging.Logger, num_tabs:int = 0):
    tabs = "\t" * num_tabs
    printable_string = f"{tabs}{logger}"

    handlers = logger.handlers
    if handlers:
        # Collect stringified handlers
        list_stringified_handlers = [f"{tabs}\t{handler}" for handler in handlers]
        stringified_handlers = _stringify_list(list_stringified_handlers, num_tabs=num_tabs+1)
        printable_string += f"\n{stringified_handlers}"
    # print(f"got handlers: {handlers}")

    return printable_string




from typing import Dict, Optional, List
from pathlib import Path

PATH_LOOKUPS = [
    {
        "local": Path("/assets"),
        "remote": Path(
            "/media/storage/Files/Dayne/WebDev/Websites/proj_ceonstock/assets"
        ),
    },
    {
        "local": Path("/storage"),
        "remote": Path(
            "/media/storage/Files/Dayne/WebDev/Websites/proj_ceonstock/local_storage"
        ),
    },
]


class PathResolver:
    def __init__(self, path_lookups: List[Dict[str, Path]]):
        self.path_lookups = path_lookups

    def local_to_remote(self, local_path: Path) -> Optional[Path]:
        for path in PATH_LOOKUPS:
            try:
                return Path(path["remote"] / local_path.relative_to(path["local"]))
            except ValueError:  # local_path is not a child of the tested path.
                pass

    def remote_to_local(self, remote_path: Path) -> Optional[Path]:
        for path in PATH_LOOKUPS:
            try:
                return Path(path["local"] / remote_path.relative_to(path["remote"]))
            except ValueError:  # local_path is not a child of the tested path.
                pass



import logging
import math
from dataclasses import dataclass
from typing import Union, Dict
from typing import Optional
from uuid import uuid4
from copy import deepcopy
from pathlib import Path

from ceonstock import config as ceonstock_config
from ceonstock import CstockProject
from ceonstock import CstockProjectInput

# from ceonstock import CstockRenderTask
# from ceonstock import CstockRenderAppType
from ceonstock import CstockJob
from ceonstock import CstockJobInput

# from ceonstock.core.render_task import CstockRenderTaskAppSettingsHou
from ceonstock.job_maker import CstockJobGenerator

from ceon_render import AppRenderJob, CeonRenderJobSettings
from ceon_render import CeonRenderPipelineJob
from ceon_render import render_apps
from ceon_render import render_pipeline as crp

from app.job_manager.job_manager import JobManager

logger = logging.getLogger(__name__)


@dataclass
class Frames:
    start: int
    end: int
    increment: int = 1

    @classmethod
    def from_string(cls, frames_str: str):
        frames_list = frames_str.split(" ")
        valid_lengths = [2, 3]
        if len(frames_list) not in valid_lengths:
            raise Exception(
                f"Invalid length '{len(frames_list)}' after splitting string {frames_str}"
            )
        start = int(frames_list[0])
        end = int(frames_list[1])
        if len(frames_list) == 3:
            increment = int(frames_list[2])
            return Frames(start=start, end=end, increment=increment)
        else:  # Increment is not present in the string
            return Frames(start=start, end=end)

    def num_frames(self) -> int:
        return (self.end - self.start) + 1

    def __post_init__(self):
        self.start = int(self.start)
        self.end = int(self.end)
        self.increment = int(self.increment)
        if self.start > self.end:
            raise Exception("Start frame cannot be great than end frame")


def _get_frames_for_project(ceonstock_project: CstockProject):
    """Take the min/max between all render tasks"""

    def _get_frames_for_task(
        ceon_render_job: CeonRenderPipelineJob,
    ) -> Optional[Frames]:
        """Get frames for each task. Return None for tasks that do not have frames"""
        file_dirs = {"project": "TODOsetprojectpathfor _get_frames_for_task"}
        app_type = ceon_render_job.app_type
        # CeonRenderJobSettings(frame_dimension=())
        if app_type.lower() == "hou":
            # app_job: render_apps.AppRenderJobHou = (
            #     crp.pipeline_job_to_concrete_job(
            #         ceonstock_project.render_pipeline,
            #         ceon_render_job,
            #         file_reference_dirs=file_dirs,
            #         render_settings=render_job_settings
            #     )
            # )
            # app_render_settings = ceon_render_job.app_render_settings
            # frames_str = app_render_settings.frames
            hou_frames = ceon_render_job.app_render_settings["frames"]
            # TODO remove string frame format, use tuples instead.
            # Currently converting for backwards compatibility. Can probably remove Frames class.
            # frames_str = f"{hou_frames[0]} {frames[1]}"
            logger.info("Got frames for hou task in _get_frames_for_task.")
            return Frames.from_string(hou_frames)
        if app_type.lower() == "ffmpeg":
            logger.warning(
                "Could not determine frames for ffmpeg task (not implemented). Returning None"
            )
            return None
        raise Exception(f"Could not determine frames for app_type: {app_type}")

    # render_settings = job_manager._create_render_settings_for_job()
    all_task_frames = [
        _get_frames_for_task(render_task)
        for render_task in ceonstock_project.render_pipeline.pipeline_jobs
    ]
    valid_task_frames = [
        task_frames for task_frames in all_task_frames if task_frames != None
    ]

    min_frame = min([frames.start for frames in valid_task_frames])
    max_frame = max([frames.end for frames in valid_task_frames])
    project_frames = Frames(start=min_frame, end=max_frame)
    return project_frames


def _get_num_entries_for_current_job(
    i: int, *, min_entries: int, max_entries: int
) -> int:
    """
    Logic to determine the number of entries to include in the job.
    i: represents the current job number within this particular job batch.
    For example, if we are generating 3 example jobs (for the same project_input), i will range from 0-2
    """
    diff = abs(max_entries - min_entries)
    middle_entries = math.floor(min_entries + (diff / 2))
    logger.debug(f"Got num_entries middle value: {middle_entries=}")
    desired = [min_entries, middle_entries, max_entries]
    default = middle_entries
    try:
        return desired[i]
    except IndexError:
        return default


def _get_desired_num_examples(project_input: CstockProjectInput):
    default = 3  # For cases that are TODO
    if project_input.web_input_type == project_input.web_input_types.DROPDOWN:
        num_dropdown_options = len(_list_dropdown_values(project_input))
        logger.debug(f"Detected {num_dropdown_options=}")
        return num_dropdown_options

    if project_input.job_input_type == project_input.job_input_types.BOOL:
        # Currently assumes one bool value. TODO handle multiple values elegantly
        # Ideal setup:
        #     All off
        #     All on
        #     Alternate 10
        #     Alternate 01
        return 2
    logger.warning(
        "TODO: Get number of generation dynamically (currently defaults to 3 for non dropdown/bools)"
    )
    # Probably want to infer based on number of 'filesets' in the input files directory.
    # E.g. admin can make folders:
    #      ImgColorTest
    #         - imgColorDemo.jpg
    #      ImgPhotos
    #         - photo1.jpg
    #         - photo2.jpg
    #         - photo3.jpg
    #      ImgDigitalArtwork
    #         - photo1.jpg
    #         - photo2.jpg
    #         - photo3.jpg
    # This would generate three examples, where each one sources from the relevant image set.
    return default


def create_jobs_for_project_input_example_video(
    ceonstock_project: CstockProject,
    *,
    project_input_name: str,
    baseline_job: CstockJob,
    job_generator: CstockJobGenerator,
) -> list[CstockJob]:
    """
    Takes a baseline job and generates various variations with different values for the target project_input_name
    job_generator: Handles the value selection logic to generate the new jobs.
    """
    # Setup
    logger.info(
        f"\nCreating jobs for project_input '{project_input_name}' example video."
    )

    target_project_input = None
    for project_input in ceonstock_project.project_inputs:
        if project_input.name == project_input_name:
            target_project_input = project_input
    if not target_project_input:
        raise Exception(f"Could not find project input named {project_input_name}")

    new_jobs = []
    num_examples_to_generate = _get_desired_num_examples(target_project_input)
    for i in range(num_examples_to_generate):
        num_entries = _get_num_entries_for_current_job(
            i,
            min_entries=target_project_input.num_entries_min,
            max_entries=target_project_input.num_entries_max,
        )
        logger.debug(f"{num_entries=}")
        logger.debug(f"{target_project_input.num_entries_min=}")
        logger.debug(f"{target_project_input.num_entries_max=}")
        new_job_input = job_generator.generate_job_input(
            target_project_input, num_entries=num_entries
        )
        logger.info(f"Created new job input: {new_job_input}")
        baseline_job_copy = create_job_copy_with_new_uuid(baseline_job)
        new_job = _replace_job_input(baseline_job_copy, new_job_input)
        new_jobs.append(new_job)

    # -----
    # Apply/calculate job frame limits
    # ----
    # Setup rendering for only a portion of the frame range.
    # Videos will be appended together.
    # TODO proper fps workflow instead of assumption
    min_job_frame_length = ceonstock_config.DEFAULT_FPS * 2
    project_frames = _get_frames_for_project(ceonstock_project)
    # Ideally, we show all examples with one pass of the video
    num_video_repeats = _calculate_required_number_of_video_repeats(
        num_jobs=len(new_jobs),
        min_job_frame_length=min_job_frame_length,
        num_project_frames=project_frames.num_frames(),
    )
    num_video_cycles = num_video_repeats + 1
    if num_video_repeats == 0:
        _split_job_frame_limits(new_jobs, project_frames)
        return new_jobs
    logger.info(
        f"Preparing frame ranges for {len(new_jobs)} jobs, spread over {num_video_cycles} video cycles"
    )
    num_jobs_per_video_cycle = math.ceil(len(new_jobs) / num_video_cycles)
    for i in range(num_video_cycles):
        start_index = num_jobs_per_video_cycle * i
        logger.debug(f"{start_index=}")
        end_index = start_index + num_jobs_per_video_cycle
        end_index = min(end_index, len(new_jobs))
        logger.debug(f"{end_index=}")
        jobs_for_cycle = new_jobs[start_index:end_index]
        logger.debug(f"{jobs_for_cycle=}")
        _split_job_frame_limits(jobs_for_cycle, project_frames)
    return new_jobs


def _split_job_frame_limits(jobs: list[CstockJob], project_frames: Frames):
    """Equally divide the available project frames among the received jobs.
    Modifies the jobs in place"""
    if not len(jobs) >= 1:
        logger.warning("Received EMPTY job of lists. No frame ranges were set.")
        return
    num_project_frames = project_frames.num_frames()
    frames_per_segment: int = math.floor(num_project_frames / len(jobs))
    logger.debug(f"Got project frames: {num_project_frames}")
    logger.debug(f"Setting job frame limits...")
    logger.debug(f"Frames per segment: {frames_per_segment}")
    logger.warning(
        "TODO: Handle missing frame in final job segment when num frames is odd."
    )
    for i, job in enumerate(jobs):
        logger.debug(
            f"job[{i}]({job.job_uuid:.5}) limit_frames: {jobs[i].limit_frames}"
        )
        frame_offset = i * frames_per_segment
        start = project_frames.start + frame_offset
        end = start + (frames_per_segment - 1)
        job.limit_frames = f"{start} {end}"
        logger.debug(
            f"job[{i}]({job.job_uuid:.5}) frame update: {jobs[i].limit_frames}"
        )


def _calculate_required_number_of_video_repeats(
    *, num_jobs: int, min_job_frame_length: int, num_project_frames: int
) -> int:
    required_total_frames = min_job_frame_length * num_jobs
    if required_total_frames <= num_project_frames:
        return 0
    num_repeats_for_all_jobs = math.floor(required_total_frames / num_project_frames)
    logger.debug(
        f"Calculated required number of video repeats for {num_jobs} jobs: {num_repeats_for_all_jobs}"
        f"\n{num_project_frames=}"
        f"\n{min_job_frame_length=}"
    )
    return num_repeats_for_all_jobs


def create_job_copy_with_new_uuid(cstock_job: CstockJob):
    """Create a copy of the current job with a new uuid"""
    job = deepcopy(cstock_job)
    job.job_uuid = str(uuid4())
    return job


def _replace_job_input(
    cstock_job: CstockJob, new_job_input: CstockJobInput
) -> CstockJob:
    """
    Replace a job input in a cstock job.
    Returns a copy.
    """
    job = deepcopy(cstock_job)
    for i, job_input in enumerate(job.job_inputs):
        if job_input.name != new_job_input.name:
            continue
        job.job_inputs[i] = new_job_input
        return job
    # The job_input name was not found
    raise Exception(
        f"Failed to replace job_input: job does not contain an input with matching name '{new_job_input.name}'."
    )


def _list_dropdown_values(project_input: CstockProjectInput) -> list:
    """
    Extract the dropdown options from the project input and return the values as a list
    """
    if not project_input.web_input_type == project_input.web_input_types.DROPDOWN:
        raise Exception(
            f"Cannot get dropdown options: Received project_input {project_input.name} is not a DROPDOWN web_input_type."
        )
    dropdown_options = project_input.type_settings["dropdown_options"]
    values = [option["value"] for option in dropdown_options]
    logger.debug(f"Got valid dropdown option values: {values}")
    return values



import inspect
from copy import deepcopy


def init_args():
    """Gets the locals() of the caller, returns without the 'self' key"""
    # TODO automatically get locals from caller so that it doesn't have to be passed by the user
    frame = inspect.currentframe()
    caller_locals = deepcopy(frame.f_back.f_locals)
    caller_locals.pop("self")
    return caller_locals



from pathlib import Path
from typing import Union, Optional, List
from ceonstock.errors import CstockError
from ceonstock.log import printify


class CstockLocalProjectError(CstockError):
    """Raised when trying to interact with a local project (files on disk)."""

    def __init__(
        self,
        project_path: Union[Path, str],
        message=None,
        errors: Optional[List] = None,
    ):
        if not message:
            message = f"Failed to load local project {project_path}"
        self.message = message
        self.errors = errors
        if errors:
            self.message += ":" + printify(errors)
        super().__init__(self.message)


class CstockLocalJobError(CstockError):
    """Raised when trying to interact with a local job (files on disk)."""

    def __init__(
        self,
        job_path: Union[Path, str],
        message=None,
        errors: Optional[List] = None,
    ):
        if not message:
            message = f"Failed to load local job {job_path}"
        self.message = message
        self.errors = errors
        if errors:
            self.message += ":" + printify(errors)
        super().__init__(self.message)



from ..core import CstockJobType
from dataclass import dataclass
from enum import Enum
import errors


class CstockRenderFPS(Enum):
    FPS_24 = 24
    FPS_25 = 24
    FPS_30 = 30
    FPS_50 = 50
    FPS_60 = 60


@dataclass
class RenderResolution:
    width: int
    height: int

    @classmethod
    def from_string(cls, string: str):
        parsed_string = string.lower().split("x")
        if len(parsed_string) == 2:
            width = int(parsed_string[0])
            height = int(parsed_string[1])
            return cls(width, height)
        raise Exception(
            f"Failed to parse string '{string}' for RenderRsolution.from_string()"
        )

    def __str__(self):
        return f"{self.width}x{self.height}"


class CstockRenderSettings():
    resolution: RenderResolution
    framerate: int
    watermarked: bool

    def __str__(self):
        # return f"<ProjInputType '{self.name}'>"
        return f"<{self.__class__.__name__}>"


DEFAULT_FRAMERATE = CstockRenderFPS.FPS_30

RENDER_PREVIEW = CstockRenderSettings(
    resolution=RenderResolution(480, 270),
    framerate=DEFAULT_FRAMERATE,
    watermarked=True
)
RENDER_STANDARD = CstockRenderSettings(
    resolution=RenderResolution(1920, 1080),
    framerate=DEFAULT_FRAMERATE,
    watermarked=False
)
RENDER_PRODUCTION = CstockRenderSettings(
    resolution=RenderResolution(3840, 2160),
    framerate=DEFAULT_FRAMERATE,
    watermarked=False
)


def default_render_settings(job_type: CstockJobType):
    CLASS_LOOKUP = {
        CstockJobType.PREVIEW: RENDER_PREVIEW,
        CstockJobType.STANDARD: RENDER_STANDARD,
        CstockJobType.PRODUCTION: RENDER_PRODUCTION
    }
    try:
        type_class = CLASS_LOOKUP[job_type]
    except KeyError:
        raise errors.UnknownJobTypeException(job_type)
    return type_class




import logging
from pathlib import Path


def init_self_logger():
    self_logger = logging.getLogger(__name__)
    # self_logger.setLevel("INFO")
    self_logger.setLevel("DEBUG")
    self_logger.info("This is self_logger before parent handler check")

    self_logger.info("Self_logger initialized!")
    return self_logger


def setup_root_logger(preserve_existing_handlers=True, level=None):
    self_logger.info("Setting up root logger...")
    root_logger = logging.getLogger()
    formatter = get_formatter("DEV")
    if root_logger.handlers:
        if preserve_existing_handlers:  # Root already has handler(s)
            for handler in root_logger.handlers:
                self_logger.info("\t...overwriting formatter only")
                handler.setFormatter(formatter)
            return
        else:
            self_logger.info("Deleting existing root handlers...")
            root_logger.handlers.clear()
    self_logger.info("root logger has no handlers!")
    self_logger.info("Adding new handler...")
    # Setup stream handler
    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    if level:
        sh.setLevel(level)
    root_logger.addHandler(sh)
    self_logger.info("Added new handler to root logger.")


class ColoredFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    green = "\x1b[32m"
    blue = "\x1b[34m"
    cyan = "\x1b[36m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    msg_format = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )

    def __init__(self, string_for_format: str):
        self.msg_format = ColoredFormatter.msg_format
        if string_for_format:
            self.msg_format = string_for_format

    def get_format(self, levelno):
        COLORS = {
            logging.DEBUG: ColoredFormatter.blue,
            logging.INFO: ColoredFormatter.grey,
            logging.WARNING: ColoredFormatter.yellow,
            logging.ERROR: ColoredFormatter.red,
            logging.CRITICAL: ColoredFormatter.bold_red,
        }
        chosen_color = COLORS.get(levelno, "")
        return chosen_color + self.msg_format + ColoredFormatter.reset

    def format(self, record):
        log_fmt = self.get_format(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_formatter(formatter_name: str) -> logging.Formatter:
    dev_formatter = ColoredFormatter(
        "%(levelname)s: [%(name)s:%(lineno)d] %(funcName)s(): %(message)s"
    )
    prod_formatter = logging.Formatter(
        "[%(name)s] %(levelname)s: %(funcName)s(): %(message)s (%(filename)s:%(lineno)d)"
    )
    formatters = {"DEV": dev_formatter, "PROD": prod_formatter}
    formatter = formatters.get(formatter_name, prod_formatter)
    return formatter


def add_stream_handler(logger):
    formatter = get_formatter("DEV")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    self_logger.info(f"Added streamHandler to: {logger.name}")


# TODO how to use logger to print messages for log creation?
# ^(without manually creating one without this function)
def print_all_loggers():
    # loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    logger_names = [name for name in logging.root.manager.loggerDict]
    """
    loggers = [
        {logging.getLogger(name): logging.getLogger(name).handlers}
        for name in logging.root.manager.loggerDict
    ]
    """
    self_logger.info(f"LOGGERS: {logger_names}")


def list_all_logger_handlers(include_null_handler=False):
    loggers = [logging.getLogger()]  # get the root logger
    loggers = loggers + [
        logging.getLogger(name) for name in logging.root.manager.loggerDict
    ]
    list_of_handlers = []
    for logger in loggers:
        for handler in logger.handlers:
            if include_null_handler or handler.__class__.__name__ != "NullHandler":
                list_of_handlers.append(f"Logger {logger} has handler {handler}")
    return list_of_handlers


def print_logger_handlers(logger_to_check):
    if logger_to_check.handlers:  # Root already has handler(s)
        self_logger.info("Found handlers!")
        for handler in logger_to_check.handlers:
            self_logger.info(f"\t{handler}")
    else:
        self_logger.info(f"Logger({logger_to_check.name}) does not have any handlers!")


# To prevent duplicate logging due to uvicorn propogation
# From https://github.com/encode/uvicorn/issues/630
def disable_uvicorn_root_logger():
    """Uvicorn adds a default handler to the root logger,
    so all logs messages are duplicated"""
    uvicorn_logger = logging.getLogger("uvicorn")
    if logging.root.handlers:
        self_logger.info(f"Got root handlers: {logging.root.handlers}")
        # Move the root logger over to the "uvicorn" logger and clear the root
        uvicorn_logger.addHandler(logging.root.handlers[0])
        logging.root.handlers = []
        uvicorn_logger.propagate = False
        self_logger.info("Disabled uvicorn root logger!")
    self_logger.info(
        "disable_uvicorn_root_logger() Did not find any root logging handlers"
    )


self_logger = init_self_logger()
if not self_logger.hasHandlers():
    setup_root_logger()
else:
    self_logger.info("This logger (or a parent) has a handler")



import os, glob, shutil, json
import logging
from dataclasses import dataclass
from typing import Optional, Union, List
from pathlib import Path

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


def copy_file(source_file: Path, dest_file: Path) -> Path:
    logger.debug(f"Copying file {source_file} to {dest_file}")
    out_dir = dest_file.parent
    if not dest_file.parent.exists():
        logger.warn(f"Created directory: {out_dir}")
        out_dir.mkdir(parents=True)
    new_file = shutil.copy(str(source_file), str(dest_file))
    return Path(new_file)


def create_directories(*paths_to_create):
    for path in paths_to_create:
        if not Path(path).exists():
            logger.info(f"Creating new folder(s): {path}")
            Path(path).mkdir(parents=True, exist_ok=True)
    """
    for path in list_of_paths
        if not os.path.exists(path):
            os.makedirs(path)
            logger.info(f"Created new folder(s): {path}")
    """


def get_files_from_folder(
    root_folder, recursive=True, raise_missing_folder=True
) -> List[Path]:
    # files = [f for f in os.listdir(folder) if os.path.isfile(os.path.join(folder, f))]
    if not Path(root_folder).is_dir():
        # TODO also confirm that the folder contains files?
        logger.warn(f"Folder does not exist: {root_folder}")
        if raise_missing_folder:
            raise Exception(f"Missing required root_folder dir: {root_folder}")
        return []
    files = glob.glob(root_folder + "/**/*", recursive=recursive)
    logger.info(f"Found files in {root_folder}: {files}")
    return files


async def save_file_to_disk(file_obj, output_file_path):
    # file_name = user_file.filename
    logger.info(f"Received an image: {file_obj}")
    with open(output_file_path, "wb+") as f:
        shutil.copyfileobj(file_obj.file, f)
    logger.info(f"Image saved to: {output_file_path}")
    return output_file_path


def zip_dir(dir_to_zip, output_zip_file_path, overwrite=False, raise_skipped_zip=False):
    # Zip project
    logger.info(f"Zipping dir {dir_to_zip} as: {output_zip_file_path}")
    output_zip_file = Path(output_zip_file_path)
    zip_dir = output_zip_file.parent
    if not zip_dir.exists():
        logger.warn(f"Dir does not exist: {zip_dir}")
        return None
    if not output_zip_file.suffix:
        logger.warn("Invalid output_zip_file_path (not a valid file path with ext)")
        return None
    if output_zip_file.exists() and not overwrite:
        msg = f"Skipping zip (already exists!): {output_zip_file_path} "
        if raise_skipped_zip:
            raise Exception(msg)
        logger.warn(msg)
        return None
    zip_type = str(output_zip_file.suffix).lstrip(".")
    logger.debug(f"Got zip type: {zip_type}")
    file_basename = output_zip_file.stem
    logger.debug(f"Got basename: {file_basename}")
    output_file_no_ext = f"{zip_dir}/{file_basename}"
    if output_zip_file.exists():
        logger.warn(f"File already exists! Overwriting: {output_zip_file_path}")
    new_zip = shutil.make_archive(output_file_no_ext, zip_type, dir_to_zip)
    logger.info(f"Successfully created zip: {new_zip}")
    return Path(new_zip)


def unzip_file(zip_file, create_folder=False, output_dir=None):
    if not os.path.isfile(zip_file):
        logger.warn(f"Project zip file not found!: {zip_file}")
        logger.debug(f"Skipping... {zip_file} (Not Found!)")
        return None
    if not output_dir:
        output_dir = Path(zip_file).parent

    if not os.path.isdir(output_dir):
        raise Exception(f"output_dir does not exist: {output_dir}")

    if create_folder:
        fldr_name = Path(zip_file).stem
        extract_to_dir = f"{output_dir}/{fldr_name}"
        if not os.path.isdir(extract_to_dir):
            os.mkdir(extract_to_dir)
    else:
        extract_to_dir = f"{output_dir}"

    logger.info(f"Unzipping: {zip_file} to {extract_to_dir}")
    shutil.unpack_archive(zip_file, extract_dir=extract_to_dir)
    logger.debug("Unzip complete")
    return extract_to_dir


def read_json_file(file_path):
    with open(file_path) as f:
        json_data = json.load(f)
    return json_data


@dataclass
class FileCheckReport:
    expected: List[str]
    found: List[str]
    missing: List[str]

    def __post_init__(self):
        if len(self.found) + len(self.missing) != len(self.expected):
            raise Exception(
                f"Failed to validate FileCheckReport. found({len(self.found)}) + missing({len(self.missing)}) != num expected({len(self.expected)})"
            )


def check_for_anomaly_files(local_dir, expected_files):
    """Returns any files in the target dir that are NOT expected"""
    local_file_list = get_files_from_folder(local_dir)
    return [file for file in local_file_list if file not in expected_files]


def check_for_expected_files(
    expected_files: Union[List[Path], List[str]]
) -> FileCheckReport:
    """Returns an object containing lists of found and missing files"""
    logger.debug(f"Got expected_files: {expected_files}")
    filepath_strings = [str(file) for file in expected_files]
    files_found = []
    files_missing = []
    for file in filepath_strings:
        if Path(file).is_file():
            files_found.append(file)
        else:
            files_missing.append(file)
    return FileCheckReport(
        expected=filepath_strings,
        found=files_found,
        missing=files_missing,
    )



from __future__ import annotations
from typing import Union, Optional, Type, List, Any, TypeVar

# from ceonstock.core.base import CstockBaseClass, CstockBaseEnum
from ceonstock.log import printify
# Use TypeVar to avoid cyclical imports

if False: # To prevent flake8 from failing on the string type-hints during CI
    from ceonstock.core.base import CstockBaseEnum
    from ceonstock.core.base import CstockBaseClass


class CstockError(Exception):
    """Base class for all Cstock exceptions"""

    pass


class CstockUnknownTypeError(CstockError):
    """Unrecognized value where an enum (or enum-matching string) was expected"""

    def __init__(
        self,
        *,
        received: Any,
        expected: Type["CstockBaseEnum"],
        message: str = "",
    ):
        if not message:
            message = f"Received invalid value for {expected.__name__}: ({type(received)}){received}"
        self.message = message
        super().__init__(self.message)


class CstockInstantiationError(CstockError):
    """Raised when a Cstock class instance could not be instantiated"""

    def __init__(
        self,
        *,
        cstock_cls: Type["CstockBaseClass"],
        errors=[],
        message: str = "",
    ):
        if not message:
            message = f"Failed to instantiate {cstock_cls.__name__}"
        self.message = message
        self.errors = errors
        if self.errors:
            self.message += f": {printify(self.errors)}"
        super().__init__(self.message)

    # def __str__(self):
    # return f"{self.message}"



from typing import List, Dict
from uuid import uuid4, UUID
from dataclasses import dataclass, field

from . import render_task_type
from . import base as cbase
from ..log import create_new_logger

logger = create_new_logger(__name__)
logger.setLevel("DEBUG")


@dataclass
class CstockRenderTask(cbase.ICstockPipelineData):
    task_name: str
    app: render_task_type.CstockRenderAppType
    app_version: str
    app_project_folder: str
    target_file: str
    app_render_settings: render_task_type.ICstockRenderAppSettings
    output: str
    dependencies: List[str] = field(default_factory=list)
    task_uuid: UUID = field(default_factory=uuid4)

    def __post_init__(self):
        if self.dependencies == None:
            self.dependencies = []
        if not self.task_uuid:
            self.task_uuid = uuid4()


# A collection of tasks that must be completed to conclude the job.
@dataclass
class CstockRenderPipeline():
    render_tasks: List[CstockRenderTask]
    outputs: Dict[str, List[str]]




from enum import Enum
from typing import Dict
from dataclasses import dataclass

from ..core.base import ICstockPipelineData


class ActionType(str, Enum):
    IMGS_FROM_DOC = "imgs_from_doc"
    IMG_CROP = "img_crop"
    IMG_RESIZE = "img_resize"
    COL_HEX_TO_RGB = "col_hex_to_rgb"  # TODO Move non-file actions elswhere?


@dataclass
class SubmissionAction(ICstockPipelineData):
    action_type: ActionType
    action_args: Dict

    @classmethod
    def from_dict(cls, dict_data):
        args = {
            "action_type": dict_data["action_type"],
            "action_args": dict_data.get("action_args", {}),
        }
        return cls(**args)

    def to_dict(self):
        new_dict = {
            "action_type": self.action_type.value,
            "action_args": self.action_args,
        }
        return new_dict
